@@ -19,10 +19,12 @@ import time
 if "." in __name__:
     from . import common
     from . import scheduler
+    from . import sqlite
     from .saxo import path as saxo_path
 else:
     import common
     import scheduler
+    import sqlite
     from saxo import path as saxo_path
 
 lock = threading.Lock()
@@ -58,7 +60,7 @@ def parse(octets):
     return match_prefix.groups(), params[0], params[1:]
 
 class ThreadSafeEnvironment(object):
-    def __init__(self, saxo, prefix, command, parameters, identified):
+    def __init__(self, saxo, prefix, command, parameters):
         self.base = saxo.base[:]
         self.config = saxo.config_cache.copy()
 
@@ -71,14 +73,10 @@ class ThreadSafeEnvironment(object):
 
         self.command = command
         self.parameters = parameters
-        self.identified = identified
 
         if command == "PRIVMSG":
             self.sender = self.parameters[0]
-            if self.identified:
-               self.text = self.parameters[1][1:]
-            else:
-               self.text = self.parameters[1]
+            self.text = self.parameters[1]
             self.private = self.sender == self.config["nick"]
             if self.private:
                 self.sender = self.nick
@@ -145,6 +143,9 @@ def socket_send(sock, flood=False):
         ...
     incoming.put(("disco_sending",))
 
+class SaxoConnectionError(Exception):
+    ...
+
 class Saxo(object):
     def __init__(self, base, opt):
         self.base = base
@@ -155,7 +156,6 @@ class Saxo(object):
         self.receiving = False
         self.sending = False
         self.user_reconnection = False
-        self.identify_msg = False
         self.links = {}
 
         self.environment_cache = os.environ.copy()
@@ -230,6 +230,15 @@ class Saxo(object):
         sys.path[:1] = []
 
     def connect(self):
+        try: self.connect_sock()
+        except Exception as err:
+           raise SaxoConnectionError(str(err))
+
+        self.first = True
+        common.thread(socket_receive, self.sock)
+        common.thread(socket_send, self.sock, "flood" in self.opt["client"])
+
+    def connect_sock(self):
         host = self.opt["server"]["host"]
         port = int(self.opt["server"]["port"])
 
@@ -244,10 +253,6 @@ class Saxo(object):
 
         debug("Connecting to %s:%s" % (host, port))
         self.sock.connect((host, port))
-        self.first = True
-
-        common.thread(socket_receive, self.sock)
-        common.thread(socket_send, self.sock, "flood" in self.opt["client"])
 
     def disconnect(self):
         outgoing.put(None) # Closes the send thread
@@ -281,7 +286,7 @@ class Saxo(object):
 
     def instruction_connected(self):
         if ":connected" in self.events:
-            irc = ThreadSafeEnvironment(self, ("", "", ""), "", [], None)
+            irc = ThreadSafeEnvironment(self, ("", "", ""), "", [])
             for function in self.events[":connected"]:
                 function(irc)
 
@@ -305,8 +310,19 @@ class Saxo(object):
                 # TODO: Close the socket?
                 ...
 
-    def instruction_identify_msg(self, ack):
-        self.identify_msg = ack
+    def instruction_instances(self):
+        our_pid = os.getpid()
+        database_filename = os.path.join(self.base, "database.sqlite3")
+        with sqlite.Database(database_filename) as db:
+            pids = [row[0] for row in db["saxo_instances"]]
+        if len(pids) == 1:
+            if pids[0] == our_pid:
+                return
+
+        debug("Found another saxo instance! %s" % pids)
+        self.send("QUIT", "Another saxo instance was detected")
+        self.disconnect()
+        os._exit(0)
 
     def instruction_join(self, channel):
         # NOTE: .visit can still be followed by .join
@@ -346,6 +362,13 @@ class Saxo(object):
         self.send("PART", channel)
 
     def instruction_ping(self):
+        # Don't ping if not connected, or connected less than a minute ago
+        if not self.receiving:
+            return
+        now = time.time()
+        if self.receiving > (now - 60):
+            return
+
         self.send("PING", self.opt["client"]["nick"])
 
         def reconnect():
@@ -368,12 +391,12 @@ class Saxo(object):
         os._exit(0)
 
     def instruction_receiving(self):
-        self.receiving = True
+        self.receiving = time.time()
         # TODO: Check that we really are connected
         # TODO: Unit test for :connected event
         incoming.put(("connected",))
 
-    def instruction_reconnect(self, close=True):
+    def instruction_reconnect(self, close=True, wait=3):
         if close:
             self.user_reconnection = True
             # Never call this from a thread, otherwise this can give an OSError
@@ -381,7 +404,7 @@ class Saxo(object):
 
         flood = "flood" in self.opt["client"]
         if not flood:
-            time.sleep(3)
+            time.sleep(wait)
         for attempt in range(7):
             if not flood:
                 if self.receiving or self.sending:
@@ -391,7 +414,10 @@ class Saxo(object):
 
         if close:
             self.user_reconnection = False
-        self.connect()
+
+        try: self.connect()
+        except SaxoConnectionError as err:
+            incoming.put(("reconnect", close, 20))
 
     def instruction_reload(self, destination=None):
         before = time.time()
@@ -403,18 +429,23 @@ class Saxo(object):
 
     def instruction_remote(self, octets):
         debug(repr(octets))
+        nick = self.opt["client"]["nick"]
         prefix, command, parameters = parse(octets)
-        identified = None
 
         if command == "PRIVMSG":
-            if self.identify_msg:
-                identified = parameters[1][0] is "+"
-                privmsg = parameters[1][1:]
-            else:
-                privmsg = parameters[1]
+            private = self.opt["client"].get("private")
+            sender = parameters[0]
+
+            if sender == nick:
+                if private == "deny":
+                    return
+                sender = prefix[0]
+            elif private == "only":
+                return
 
             pfx = self.opt["client"]["prefix"]
             length = len(pfx)
+            privmsg = parameters[1]
 
             if privmsg.startswith(pfx):
                 privmsg = privmsg[length:]
@@ -423,11 +454,7 @@ class Saxo(object):
                 else:
                     cmd, arg = privmsg, ""
 
-                sender = parameters[0]
-                if sender == self.opt["client"]["nick"]:
-                    sender = prefix[0]
-
-                self.command(prefix, sender, identified, cmd, arg)
+                self.command(prefix, sender, cmd, arg)
 
         elif command == "PONG":
             if self.discotimer is not None:
@@ -438,8 +465,7 @@ class Saxo(object):
                 except:
                     ...
 
-
-        args = (self, prefix, command, parameters, identified)
+        args = (self, prefix, command, parameters)
         irc = ThreadSafeEnvironment(*args)
         def safe(function, irc):
             try: function(irc)
@@ -471,7 +497,7 @@ class Saxo(object):
     def instruction_sending(self):
         self.sending = True
 
-    def command(self, prefix, sender, identified, cmd, arg):
+    def command(self, prefix, sender, cmd, arg):
         if ("\x00" in cmd) or (os.sep in cmd) or ("." in cmd):
             return
 
@@ -511,10 +537,6 @@ class Saxo(object):
             env = self.environment_cache.copy()
             env["SAXO_NICK"] = prefix[0]
             env["SAXO_SENDER"] = sender
-            if identified == True:
-                env["SAXO_IDENTIFIED"] = "1"
-            if identified == False:
-                env["SAXO_IDENTIFIED"] = "0"
             if sender in self.links:
                 env["SAXO_URL"] = self.links[sender]
             common.thread(process, env, path, arg)
